/*
 *    XenoStep_theme.h
 */

#ifndef __XENOSTEP_THEME_H__
#define __XENOSTEP_THEME_H__

#include <math.h>
#include <gtk/gtk.h>

#ifdef G_LOG_DOMAIN
#undef G_LOG_DOMAIN
#endif
#define G_LOG_DOMAIN "XenoStep-Theme"


/* Default Data values */
#define XENOSTEP_SHADE_MUL                  16384.0

#define XENOSTEP_DEFAULT_SHINE              1.5
#define XENOSTEP_DEFAULT_SHADE              0.6666667
#define XENOSTEP_DEFAULT_SCROLLBAR_WIDTH    15


/* Frame types */
typedef enum {
    XENOSTEP_SHADOW_NONE        = GTK_SHADOW_NONE,
    XENOSTEP_SHADOW_IN          = GTK_SHADOW_IN,
    XENOSTEP_SHADOW_OUT         = GTK_SHADOW_OUT,
    XENOSTEP_SHADOW_MENU        = XENOSTEP_SHADOW_OUT,
    XENOSTEP_SHADOW_ETCHED_IN   = GTK_SHADOW_ETCHED_IN,
    XENOSTEP_SHADOW_ETCHED_OUT  = GTK_SHADOW_ETCHED_OUT,

    XENOSTEP_SHADOW_BUTTON_IN,
    XENOSTEP_SHADOW_NONE_OUT,
    XENOSTEP_SHADOW_NONE_IN,
    XENOSTEP_SHADOW_IN_FOCUSED,
    XENOSTEP_SHADOW_THIN_OUT,
    XENOSTEP_SHADOW_THIN_IN,
    
    XENOSTEP_SHADOW_NOTE,
    XENOSTEP_SHADOW_LAST
} XenoStepShadowType;


/* Images */
#define XENOSTEP_RADIO_SIZE         13
#define XENOSTEP_CHECK_SIZE         14
#define XENOSTEP_CHECK_MENU_SIZE    8
#define XENOSTEP_RADIO_MENU_SIZE    6

typedef enum {
    XENOSTEP_IMAGE_CHECK_BUTTON_NORMAL,
    XENOSTEP_IMAGE_CHECK_BUTTON_PRELIGHT,
    XENOSTEP_IMAGE_CHECK_BUTTON_INSENSITIVE,
    XENOSTEP_IMAGE_CHECK_BUTTON_ACTIVE,
    
    XENOSTEP_IMAGE_CHECK_MENU_NORMAL,
    XENOSTEP_IMAGE_CHECK_MENU_PRELIGHT,
    XENOSTEP_IMAGE_CHECK_MENU_INSENSITIVE,

    XENOSTEP_IMAGE_RADIO_NORMAL_OUT,
    XENOSTEP_IMAGE_RADIO_PRELIGHT_OUT,
    XENOSTEP_IMAGE_RADIO_INSENSITIVE_OUT,
    XENOSTEP_IMAGE_RADIO_NORMAL_IN,
    XENOSTEP_IMAGE_RADIO_PRELIGHT_IN,
    XENOSTEP_IMAGE_RADIO_INSENSITIVE_IN,
    XENOSTEP_IMAGE_RADIO_ACTIVE_IN,
    
    XENOSTEP_IMAGE_RADIO_MENU_NORMAL,
    XENOSTEP_IMAGE_RADIO_MENU_PRELIGHT,
    XENOSTEP_IMAGE_RADIO_MENU_INSENSITIVE,

    XENOSTEP_IMAGE_NOTEBOOK_CORNER_NORMAL,
    XENOSTEP_IMAGE_NOTEBOOK_CORNER_ACTIVE,
    XENOSTEP_IMAGE_NOTEBOOK_CORNER_INSENSITIVE,

    XENOSTEP_IMAGE_DIMPLE_NORMAL,
    XENOSTEP_IMAGE_DIMPLE_PRELIGHT,
    
    XENOSTEP_IMAGE_LAST
} XenoStepImageType;

typedef struct _XenoStepImage       XenoStepImage;
typedef struct _XenoStepColorUsage  XenoStepColorUsage;

struct _XenoStepColorUsage {
    guint8 flags[5];
};

enum {
    XENOSTEP_USE_FG     = 1,
    XENOSTEP_USE_BG     = 2,
    XENOSTEP_USE_TEXT   = 4,
    XENOSTEP_USE_BASE   = 8
};

#define XENOSTEP_USE(pen, state)    ((XENOSTEP_USE_ ## pen) * 5 + state)

typedef enum {
    XENOSTEP_MASK_CHECK,
    XENOSTEP_MASK_RADIO,
    XENOSTEP_MASK_RADIO_ITEM,
    XENOSTEP_MASK_CORNER,
    XENOSTEP_MASK_DIMPLE,

    XENOSTEP_MASK_LAST
} XenoStepMaskIndex;

struct _XenoStepImage {
    const char          **xpm;
    XenoStepMaskIndex   mask;
    const guchar        *remap;
    GtkStateType        state_type:8;
    guint8              color_usage;
};


/* Configuration */
enum {
    XENOSTEP_TOKEN_FIRST    = G_TOKEN_LAST + 1,
    XENOSTEP_TOKEN_TRUE     = XENOSTEP_TOKEN_FIRST,
    XENOSTEP_TOKEN_FALSE,
    
    XENOSTEP_TOKEN_STEPPER_ENDS,
    XENOSTEP_TOKEN_STEPPER_ARROWS,
    XENOSTEP_TOKEN_STEPPER_BOX,
    XENOSTEP_TOKEN_SCROLLBAR_WIDTH,
    XENOSTEP_TOKEN_SCROLLBAR_KNOB,
    XENOSTEP_TOKEN_PANED_KNOB,
    XENOSTEP_TOKEN_HANDLE_KNOB,
    XENOSTEP_TOKEN_BUTTONDEF_SHADOW,
    XENOSTEP_TOKEN_TOOLTIP_SHADOW,
    
    XENOSTEP_TOKEN_THIN,
    XENOSTEP_TOKEN_SHINE,
    XENOSTEP_TOKEN_SHADE,
    
    XENOSTEP_TOKEN_NONE,
    XENOSTEP_TOKEN_DIMPLE,
    XENOSTEP_TOKEN_LINES,
    XENOSTEP_TOKEN_HOLES,
    XENOSTEP_TOKEN_BUDS,

    XENOSTEP_TOKEN_NORMAL,
    XENOSTEP_TOKEN_PRELIGHT,
    XENOSTEP_TOKEN_ACTIVE,
    XENOSTEP_TOKEN_SELECTED,
    XENOSTEP_TOKEN_INSENSITIVE,
    
    XENOSTEP_TOKEN_SHADOW_IN,
    XENOSTEP_TOKEN_SHADOW_OUT,
    XENOSTEP_TOKEN_SHADOW_MENU,
    XENOSTEP_TOKEN_SHADOW_ETCHED_IN,
    XENOSTEP_TOKEN_SHADOW_ETCHED_OUT,
    XENOSTEP_TOKEN_SHADOW_BUTTON_IN,
    XENOSTEP_TOKEN_SHADOW_NONE_OUT,
    XENOSTEP_TOKEN_SHADOW_NONE_IN,
    XENOSTEP_TOKEN_SHADOW_IN_FOCUSED,
    XENOSTEP_TOKEN_SHADOW_THIN_OUT,
    XENOSTEP_TOKEN_SHADOW_THIN_IN,
    XENOSTEP_TOKEN_SHADOW_NOTE,
    
    XENOSTEP_TOKEN_LAST
};

enum {
    XENOSTEP_DATA_STEPPER_ENDS      = 0x01,
    XENOSTEP_DATA_STEPPER_ARROWS    = 0x02,
    XENOSTEP_DATA_STEPPER_BOX       = 0x04,
    XENOSTEP_DATA_SCROLLBAR_WIDTH   = 0x08,
    XENOSTEP_DATA_SCROLLBAR_KNOB    = 0x10,
    XENOSTEP_DATA_PANED_KNOB        = 0x20,
    XENOSTEP_DATA_HANDLE_KNOB       = 0x40,
    XENOSTEP_DATA_BUTTONDEF_SHADOW  = 0x80,
    XENOSTEP_DATA_TOOLTIP_SHADOW    = 0xC0,
    XENOSTEP_DATA_THIN              = 0xF0,
    XENOSTEP_DATA_SHADE             = 0x00106,
    XENOSTEP_DATA_SHINE             = 0x04006
};

typedef enum {
    XENOSTEP_KNOB_NONE,
    XENOSTEP_KNOB_DIMPLE,
    XENOSTEP_KNOB_LINES,
    XENOSTEP_KNOB_HOLES,
    XENOSTEP_KNOB_BUDS
} XenoStepKnobStyle;

typedef struct {
    gint16 ref_count;

    gboolean thin           : 1;
    gboolean stepper_ends   : 1;
    gboolean stepper_arrows : 1;
    gboolean stepper_box    : 1;
    
    XenoStepKnobStyle   scrollbar_knob;
    XenoStepKnobStyle   paned_knob;
    XenoStepKnobStyle   handle_knob;
    XenoStepShadowType  buttondef_shadow;
    XenoStepShadowType  tooltip_shadow;

    guint32 config;
    
    guint8  scrollbar_width;
    guint8  npixmaps;

    guint8  dirty;
    guint16 shine[5];
    guint16 shade[5];

    GdkPixmap *pixmaps[XENOSTEP_IMAGE_LAST];
} XenoStepThemeData;


/* in xenostep_theme_color.c */
typedef struct _XenoStepRGB {
    gdouble r, g, b;
} XenoStepRGB;

typedef struct _XenoStepHSL {
    gdouble h, s, l;
} XenoStepHSL;

enum {
    /* colors in the style */
    XENOSTEP_FG,    XENOSTEP_BG,
    XENOSTEP_LIGHT, XENOSTEP_DARK, XENOSTEP_MID,
    XENOSTEP_TEXT,  XENOSTEP_BASE,
    XENOSTEP_WHITE, XENOSTEP_BLACK,
    
    /* conditional */       
    XENOSTEP_THIN,
    
    /* change hue only */
    XENOSTEP_HUE,
    
    /* blends */
    XENOSTEP_BI,
    XENOSTEP_TRI,
    XENOSTEP_QUAD,
    
    /* static */
    XENOSTEP_USER
};

enum {
    XENOSTEP_NORMAL         = 0,
    XENOSTEP_ACTIVE         = 1,
    XENOSTEP_PRELIGHT       = 2,
    XENOSTEP_SELECTED       = 3,
    XENOSTEP_INSENSITIVE    = 4,
    XENOSTEP_CURRENT        = 5,
    XENOSTEP_PARENT         = 0x10
};


/* xenostep_theme_color.c */
void        xenostep_rgb_to_hsl (const XenoStepRGB *src, XenoStepHSL *dst);
void        xenostep_hsl_to_rgb (const XenoStepHSL *src, XenoStepRGB *dst);
        
void        xenostep_color_to_hsl (const GdkColor *src, XenoStepHSL *dst);
void        xenostep_hsl_to_color (const XenoStepHSL *src, GdkColor *dst);
        
void        xenostep_color_to_rgb (const GdkColor *src, XenoStepRGB *dst);
void        xenostep_rgb_to_color (const XenoStepRGB *src, GdkColor *dst);

gboolean    xenostep_color_equal (const GdkColor *a, const GdkColor *b);

void        xenostep_shade_color (const GdkColor *src, GdkColor *dst, gdouble k);
void        xenostep_blend_color (const GdkColor *a, const GdkColor *b, gdouble k, GdkColor *dst);


/* xenostep_theme_main.c */
void        xenostep_xpm_remap (const GtkStyle *style, const GtkStyle *style2,
                                GtkStateType state_type, const guchar * stream,
                                char **xpm);

#define xenostep_mask_get(i) \
    xenostep_masks[xenostep_images[i].mask]

GdkPixmap * xenostep_pixmap_get (GdkWindow *window, GtkStyle *style,
                                 const GtkStyle *style2,
                                 XenoStepImageType image_type);

gboolean    xenostep_pixmap_use_colors (XenoStepImageType image_type,
                                        guint8 usage);

char **     xenostep_xpm_copy (const char **xpm);
void        xenostep_xpm_free (char **xpm);


/* xenostep_theme_patches.c */
void        xenostep_patches_install ();
void        xenostep_patches_remove ();


/* Exported data structures */
extern      GtkStyleClass   xenostep_class;
extern      GtkStyleClass   xenostep_thin_class;

extern      GdkBitmap       * xenostep_masks[XENOSTEP_MASK_LAST];
extern      XenoStepImage   xenostep_images[XENOSTEP_IMAGE_LAST];

#endif

